# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import report_import
report_import("zef.core.graph_slice")


from ._core import *
from .abstract_raes import *
from .internals import EternalUID
from .VT import *

class GraphSlice_:
    def __init__(self, *args):
        # TODO: Temp imports, going to move op definitions around
        from ._ops import to_ezefref
        """
        Construct a GraphSlice, aka reference frame.

        Signature:
            TX -> GraphSlice
            (Time, Graph) -> GraphSlice
            (Graph, Time) -> GraphSlice
        """

        # ZefRef[TX] -> GraphSlice and EZefRef[TX] -> GraphSlice
        if len(args) == 1:
            z = args[0]
            from .VT import TX
            if not isinstance(z, TX):
                raise TypeError(f'When calling the GraphSlice constructor with a single arguments, this has to be a TX. Called with: args={args}')
            if isinstance(z, AtomClass):
                from .atom import _get_ref_pointer
                z = _get_ref_pointer(z)
                if z is None:
                    raise TypeError(f"Atom did not have a BlobPtr contained in it: {args}.")
            self.tx = to_ezefref(z)
            # Hold a reference to the graph open
            self._g = Graph(self.tx)
            return

        if len(args) == 2:
            if {type(args[0]), type(args[1])} != {Graph, Time}:
                raise TypeError(f'When calling the GraphSlice constructor with two args, this has to be with types (Time, Graph). Called with args={args}')

            t, g = args if isinstance(args[0], Time) else args[::-1]
            if t > ops.now():
                raise RuntimeError('One cannot ask for a GraphSlice at a fixed time in the future.')
            txs = g | ops.all[zefdb.TX] | ops.collect
            if t < ops.collect(txs | ops.first | ops.time):
                raise RuntimeError(f'In trying to determine a GraphSlice given a time: provided time {t} is before the graph instantiation time.')
            
            last_tx = txs | last
            self.tx = (
                last_tx                                 # treat the case of the 
                if t >= time(last_tx) else
                txs | filter[lambda x: time(x) >= t] | ops.first | ops.collect
            )
            # Hold a reference to the graph open
            self._g = Graph(self.tx)
            return

        raise TypeError(f'The GraphSlice constructor can only be called with one or two args. Called with args={args}')
        
        
    def __repr__(self):
        from ._ops import uid, graph_slice_index
        return f"GraphSlice(graph_uid='{str(uid(Graph(self.tx)))}', index={graph_slice_index(self)})"

    def __eq__(self, other):        
        return (
            self.tx == other.tx 
            if isinstance(other, GraphSlice)
            else False
            )
            
    def __int__(self):
        """
        Returns the time slice as an integer
        """
        return self | graph_slice_index | collect

    def __hash__(self):
        return hash(self.tx)

    def __getitem__(self, thing):
        from ._ops import to_frame, uid, collect, to_delegate, exists_at
        from . import internals

        from .VT import Delegate
        if isinstance(thing, Delegate):
            # In case thing is a BlobPtr, convert it to DelegateRef first
            d = to_delegate(thing)
            maybe_z = to_delegate(d, self)
            if maybe_z is None:
                raise KeyError(f"Delegate {thing} not present in this timeslice")
            return maybe_z

        from .VT import Val
        if isinstance(thing, Val):
            val = internals.val_as_serialized_if_necessary(thing)
            maybe_z = Graph(self.tx).get_value_node(val)
            if maybe_z is None:
                raise KeyError(f"ValueNode {thing} doesn't exist on graph") 
            if not maybe_z | exists_at[self] | collect:
                raise KeyError(f"ValueNode {thing} isn't alive in this timeslice") 
            return maybe_z | to_frame[self] | collect
        
        g = Graph(self.tx)
        ezr = g[thing]
        # We magically transform any FOREIGN_ENTITY_NODE accesses to the real RAEs.
        # Accessing the low-level BTs can only be done through traversals
        if BT(ezr) in [BT.FOREIGN_ENTITY_NODE, BT.FOREIGN_ATTRIBUTE_ENTITY_NODE, BT.FOREIGN_RELATION_EDGE]:
            res = get_instance_rae(uid(thing), self)
            if res is None:
                raise KeyError("RAE doesn't have an alive instance in this timeslice")
            return res

        else:
            return ezr | to_frame[GraphSlice(self.tx)] | collect

    def __contains__(self, thing):
        from ._ops import exists_at, uid, collect, to_delegate
        from . import internals
        from .rae_type_definitions import RAE

        if isinstance(thing, RAE):
            return get_instance_rae(origin_uid(thing), self) is not None

        from .VT import Delegate
        if isinstance(thing, Delegate):
            # In case thing is a BlobPtr, convert it to DelegateRef first
            d = to_delegate(thing)
            maybe_z = to_delegate(d, self)
            return maybe_z is not None

        from .VT import Val
        if isinstance(thing, Val):
            val = internals.val_as_serialized_if_necessary(thing)
            maybe_z = Graph(self.tx).get_value_node(val)
            if maybe_z is None:
                return False
            return maybe_z | exists_at[self] | collect
            

        g = Graph(self.tx)
        if isinstance(thing, EternalUID):
            z = get_instance_rae(thing, self)
            return z is not None

        if thing not in g:
            return False
        z = g[thing]

        return z | exists_at[GraphSlice(self.tx)] | collect

from .VT import make_VT
GraphSlice = make_VT("GraphSlice", pytype=GraphSlice_)


def get_instance_rae(origin_uid: EternalUID, gs: GraphSlice, allow_tombstone=False)->ZefRef:
    """
    Returns the instance of a foreign rae in the given slice. It could be that
    the node asked for has its origin on this graph (the original rae may still
    be alive or it may be terminated)

    Args:
        origin_uid (EternalUID): the uid of the origin rae we are looking for
        g (Graph): on which graph are we looking?

    Returns:
        ZefRef: this graph knows about this: found instance
        None: this graph knows nothing about this RAE
    """
    # TODO: Temporary imports while ops are moving around
    from ._ops import map, target, exists_at, filter, collect, only, to_frame, Ins
    g = Graph(gs.tx)
    if origin_uid not in g:
        return None

    zz = g[origin_uid]
    if BT(zz) in {BT.FOREIGN_ENTITY_NODE, BT.FOREIGN_ATTRIBUTE_ENTITY_NODE, BT.FOREIGN_RELATION_EDGE}:
        z_candidates = zz | Ins[BT.ORIGIN_RAE_EDGE] | map[target] | collect
        if not allow_tombstone:
            z_candidates = z_candidates | filter[exists_at[gs]] | collect
        if len(z_candidates) > 1:
            raise RuntimeError(f"Error: More than one instance alive found for RAE with origin uid {origin_uid}")
        elif len(z_candidates) == 1:
            from . import _ops
            return z_candidates | only | to_frame[gs][_ops.allow_tombstone] | collect
        else:
            return None     # no instance alive at the moment
        
    elif BT(zz) in {BT.ENTITY_NODE, BT.ATTRIBUTE_ENTITY_NODE, BT.RELATION_EDGE, BT.TX_EVENT_NODE, BT.ROOT_NODE}:
        if allow_tombstone:
            from . import _ops
            return zz | to_frame[gs][_ops.allow_tombstone] | collect
        else:
            return zz | to_frame[gs] | collect
    else:
        raise RuntimeError(f"Unexpected option in get_instance_rae: {zz}")


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# This file implements Cond[TestVT][SecondaryVT]
#
# It is equivalent to an implication, that if the first applies, so must the
# second.
#
# Alternatively, *either* TestVT doesn't apply, or SecondaryVT must apply, i.e.
# ~TestVT | SecondaryVT.
#
# The rest of this code is just a convoluted way to express:
# Cond[X][Y] = ~X | Y

from . import make_VT, ValueType

from .helpers import remove_names, absorbed

def cond_validation(typ):
    abs = remove_names(absorbed(typ))
    if len(abs) != 2:
        raise Exception(f"Cond needs exactly one absorbed argument, have {abs}")
    for x in abs:
        if not isinstance(x, ValueType):
            raise Exception(f"Cond requires ValueTypes. Found {x}")
    return True

def cond_is_a(x, typ):
    assert cond_validation(typ)
    
    test,follow_up = remove_names(absorbed(typ))

    # This test is like a logic implication so we can replace it with:
    equivalent = ~test | follow_up

    return isinstance(x, equivalent)

Cond = make_VT("Cond",
               is_a_func=cond_is_a)


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import report_import
report_import("zef.core.graph_additions.command_ordering")

from .common import *

def order_level1_commands(commands: List[PleaseCommandLevel1], gs: GraphSlice):
    # Form dependency graph using WishIDs and Refs
    #
    # Take partial ordering from graph, using order of original commands to decide ties.

    def simple_key(cmd):
        if isinstance(cmd, PleaseInstantiate):
            if isinstance(cmd.atom, PleaseInstantiateEntity | PleaseInstantiateAttributeEntity):
                return 0
            elif isinstance(cmd.atom, PleaseInstantiateRelation):
                return 1
        elif isinstance(cmd, PleaseAssign):
            return 10
        return 999

    commands = commands | sort[simple_key] | collect


    # TODO: Build a dependency graph
    if True:
        new_commands = []

        from collections import defaultdict
        cmd_mapping = {}
        ready_cmds = []
        dependencies = {}
        dependents = defaultdict(lambda: [])

        # Build graph
        for cmd in commands:
            name,deps = (cmd,gs) | match_rules[[
                *command_dep_rules,
                (Any, not_implemented_error["Command unknown for dependency ordering"]),
            ]] | collect

            # Special case for no id - can't have any dependents
            if name is None:
                assert len(deps) == 0
                new_commands += [cmd]
                continue

            cmd_mapping[name] = cmd

            if len(deps) == 0:
                ready_cmds += [name]
            else:
                dependencies[name] = deps
                for dep in deps:
                    dependents[dep] += [name]

        # Work through commands that have their deps done
        while len(ready_cmds) > 0:
            name = ready_cmds.pop(0)
            new_commands += [cmd_mapping[name]]

            # Free up other commands that dependended on this one
            for other in dependents[name]:
                dependencies[other].remove(name)
                if len(dependencies[other]) == 0:
                    ready_cmds += [other]

        # We only went through all "ready" commands. There might have been some
        # that never became ready in an invalid wish case.
        if len(new_commands) != len(commands):
            reverse_mapping = {b:a for a,b in cmd_mapping.items()}
            print()
            print("======")
            for cmd in commands:
                if cmd in new_commands:
                    continue
                name = reverse_mapping[cmd]
                print("----")
                print(name)
                print(cmd)
                print("--")
                for dep in dependencies[name]:
                    print(dep)
                
            print()
            raise Exception("Unable to order all commands")

        commands = new_commands
            
    assert are_commands_ordered(commands, gs)
    # TODO: Temporary for testing
    return commands

def are_commands_ordered(commands: List[PleaseCommandLevel1], gs: GraphSlice):
    # Instead of generating a list, this should really just check that the
    # dependencies satisfy their ordering requirements.

    # Build a dependency graph
    cmd_mapping = {}
    dependencies = {}

    for ind,cmd in enumerate(commands):
        name,deps = (cmd,gs) | match_rules[[
            *command_dep_rules,
            (Any, not_implemented_error["Command unknown for dependency ordering"]),
        ]] | collect

        # Special case for no id - can't have any dependents
        if name is None:
            continue

        cmd_mapping[name] = ind

        if len(deps) > 0:
            dependencies[name] = deps

    # Check the order in the graph
    for name,deps in dependencies.items():
        for dep in deps:
            if cmd_mapping[dep] > cmd_mapping[name]:
                return False

    return True




def deps_instantiate(cmd, gs):
    if isinstance(cmd.atom, PleaseInstantiateRelation):
        deps = []
        if not isinstance(cmd.atom["source"], EternalUID | DelegateRef | WrappedValue) or cmd.atom["source"] not in gs:
            deps += [("Instantiate", cmd.atom['source'])]
        if not isinstance(cmd.atom["target"], EternalUID | DelegateRef | WrappedValue) or cmd.atom["target"] not in gs:
            deps += [("Instantiate", cmd.atom['target'])]
    else:
        deps = []
        
    if isinstance(cmd.atom, PleaseInstantiateDelegate | PleaseInstantiateValueNode):
        name = ("Instantiate", cmd.atom)
    elif "origin_uid" in cmd:
        name = ("Instantiate", cmd.origin_uid)
    elif "internal_ids" in cmd:
        name = ("Instantiate", single(cmd.internal_ids))
    else:
        name = None
    return name,deps

def deps_terminate(cmd, gs):
    name = ("Terminate", cmd.target)
    deps = []
    return name, deps

def deps_assign(cmd, gs):
    name = ("Assign", cmd.target)
    deps = []
    if not isinstance(cmd.target, EternalUID) or cmd.target not in gs:
        deps += [("Instantiate", cmd.target)]
    return name,deps

command_dep_rules = [
    (PleaseInstantiate, deps_instantiate),
    (PleaseTerminate, deps_terminate),
    (PleaseAssign, deps_assign),
]


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import report_import
report_import("zef.core.graph_additions.wish_translation")

from .types import *
from .common import *


SetFieldAction = LazyValue[Any][ZefOp[set_field]]
# TODO: More stuff here
SetFieldActionValid = LazyValue[AtomClass | AllIDs][ZefOp[set_field]]

##########################################
# * Level 1 generation
#----------------------------------------

class CouldNotProcess: pass

WishMapping = Dict[WishID][EZefRef|WishID]
# CmdDispatchContext = UserValueType("Level1CmdDispatchContext",
#                                    Dict,
#                                    # Pattern[{"mapping": WishMapping, "gen_id_state": GenIDState}]
#                                    Pattern[{"gen_id_state": GenIDState}]
#                                    )
CmdDispatchContext = Pattern[{"gen_id_state": GenIDState}]

DispatchOutput = SetOf(CouldNotProcess) | Tuple[List[PleaseCommandLevel1], List[PleaseCommandLevel2], CmdDispatchContext]

# Rules should follow the signature below, but we can't specify that just yet.
# TranslationFunc = Func[PleaseCommandLevel2, CmdDispatchContext][DispatchOutput]
TranslationFunc = Any


from .command_preparation import dispatch_preparation
from .command_multi_rules import relabel_cmds, cull_unnecessary_cmds, distinguish_cmd, resolved_variables_from_aliases
from .command_recombination import recombination
from .command_ordering import order_level1_commands




# Really want to use Subtype[PleaseCommandLevel2] instead of ValueType
def generate_level1_commands(commands: List[PleaseCommandLevel2], gs: GraphSlice) -> Level1CommandInfo:
    # Blind generation, using no culling at this stage, merely resolving PleaseRun commands

    # context = CmdDispatchContext({
    #     "gen_id_state": generate_initial_state("lvl1"),
    # })
    context = {"gen_id_state": generate_initial_state("lvl1")}
    
    # We have to have multiple passes at the commands as each command can slowly
    # describe different wish ids.

    # Resolve the arbitrary please run commands and object notation
    todo = list(commands)
    prepared_cmds = []
    
    # print("*** Start preparation loop")
    while len(todo) > 0:
        cmd = todo.pop(0)
       
        ready_cmds, new_todo, context = dispatch_preparation(cmd,gs,context)
        # print()
        # print("===")
        # print(ready_cmds)
        # print(new_todo)
        # print()

        prepared_cmds += ready_cmds
        todo += new_todo

    # print()
    # print("AFTER PREPARATION RUN")
    # for cmd in prepared_cmds:
    #     print(cmd)
    # print()

    # Now we know all after_please_run commands are level1 commands, but we need
    # to include references where there are relations being affected, as these
    # rely on their source/target remaining alive. This is also where we would
    # include dependencies on other nodes potentially terminated via hostage
    # flags.
    out_cmds = []

    def add_must_live_cmds(src_or_trg):
        if isinstance(src_or_trg, WishID):
            id = src_or_trg
        elif isinstance(src_or_trg, Delegate):
            return
        elif isinstance(src_or_trg, WrappedValue):
            return
        elif isinstance(src_or_trg, Relation):
            add_must_live_cmds(source(src_or_trg))
            add_must_live_cmds(target(src_or_trg))
            id = origin_uid(src_or_trg)
        elif isinstance(src_or_trg, AtomClass):
            id = origin_uid(src_or_trg)
        elif isinstance(src_or_trg, EternalUID):
            id = src_or_trg
            z = find_rae_in_target(id, gs)
            if isinstance(z, Relation):
                add_must_live_cmds(source(z))
                add_must_live_cmds(target(z))
        elif isinstance(src_or_trg, BlobPtr):
            id = origin_uid(src_or_trg)
            if isinstance(src_or_trg, Relation):
                add_must_live_cmds(source(src_or_trg))
                add_must_live_cmds(target(src_or_trg))
        else:
            raise Exception(f"Shouldn't get here: {src_or_trg}")
        nonlocal out_cmds
        out_cmds += [PleaseMustLive({"target": id})]
            
    for cmd in prepared_cmds:
        if isinstance(cmd, PleaseInstantiate) and isinstance(cmd.atom, PleaseInstantiateRelation):
            add_must_live_cmds(cmd.atom["source"])
            add_must_live_cmds(cmd.atom["target"])
        elif isinstance(cmd, PleaseTag):
            if isinstance(cmd.target, RelationRef):
                add_must_live_cmds(source(cmd.target))
                add_must_live_cmds(target(cmd.target)) 
        elif isinstance(cmd, PleaseInstantiate | PleaseAssign | PleaseAlias | PleaseTerminate | PleaseMustLive | PleaseBeSource | PleaseBeTarget):
            pass
        # elif hostage detection things...
        else:
            raise Exception(f"Shouldn't get here in realising implicit constraints: {cmd}")

        out_cmds += [cmd]

    # print()
    # print("OUT CMDS from implicit constraints")
    # for cmd in out_cmds:
    #     print(cmd)
    # print()

    # Now we know all out_cmds are level1 commands, including all possible
    # dependents (e.g. relation source/target references). But we don't know if
    # these are valid or can be simplified.

    resolved_variables = {}
    cur_cmds = out_cmds
    while True:
        new_cmds,simplify_aliases,did_something = validate_and_simplify_lvl1_cmds(cur_cmds)
        if not did_something:
            break
        cur_cmds = relabel_cmds(new_cmds, simplify_aliases)
        resolved_variables = resolved_variables | merge[resolved_variables_from_aliases(simplify_aliases)] | collect
        # print()
        # print("NAMED OUTPUT CMDS")
        # for cmd in cur_cmds:
        #     print(cmd)
        # print()

    named_output_cmds = cur_cmds


    culled_cmds,cull_aliases = cull_unnecessary_cmds(named_output_cmds, gs)
    # TODO: Apply aliases
    culled_cmds = relabel_cmds(culled_cmds, cull_aliases)
    resolved_variables = resolved_variables | merge[resolved_variables_from_aliases(cull_aliases)] | collect

    # print()
    # print("CULLED CMDS")
    # for cmd in culled_cmds:
    #     print(cmd)
    # print("---")

    ordered_cmds = order_level1_commands(culled_cmds, gs)

    # print()
    # print("ORDERED CMDS")
    # for cmd in ordered_cmds:
    #     print(cmd)
    # print("---")

    # print()
    # print("RESOLVED VARIABLES:")
    # for a,b in resolved_variables.items():
    #     print(a,b)
    # print("---")

    return Level1CommandInfo(gs=gs,
                             cmds=ordered_cmds,
                             resolved_variables=resolved_variables)


def validate_and_simplify_lvl1_cmds(cmds_in):
    # We need to reduce the ambiguity on naming of commands, so we build up a
    # bipartite graph of which names/cmds refer to each other.
    #
    # Note that relations only need to refer to their own relation, not their
    # source/target. Any validation tracking has been taken care of with the
    # PleaseMustLive on the source/targets already and any relabelling will
    # happen after we first handle the simplification process.

    # WAIT! What happens if two relations are being merged in and they have
    # different source/targets. We can reduce the distinguishable part to the
    # uid/internal id only, where internal id wins (just as for entities). But
    # in recombination, checks will be made on the source/target and an
    # invalidation error will be throw at that point.

    did_something = False

    from collections import defaultdict
    name_mapping = defaultdict(lambda: set())
    cmd_mapping = defaultdict(lambda: set())

    for cmd in cmds_in:
        cmd_distinguish,names = distinguish_cmd(cmd)
        assert len(names) > 0, f"Command without any names would otherwise be silently ignored! {cmd}"
        if cmd != cmd_distinguish:
            did_something = True
        for name in names:
            name_mapping[name].add(cmd_distinguish)
        cmd_mapping[cmd_distinguish].update(names)
            

    # print()
    # print("NAME MAPPING")
    # for name,cmds in name_mapping.items():
    #     print(name)
    #     for cmd in cmds:
    #         print("  -", cmd)
    # print("---")

    # Find any connected components to simplify multiple names.
    relabelled_names = {}
    new_name_mapping = {}

    todo_names = set(name_mapping.keys())
    components = []
    while len(todo_names) > 0:
        this_todo_names = {todo_names.pop()}
        comp_names = set()
        while len(this_todo_names) > 0:
            name = this_todo_names.pop()
            comp_names.add(name)

            hop2_names = {x for cmd in name_mapping[name] for x in cmd_mapping[cmd]}
            assert len(hop2_names) > 0
            hop2_names = hop2_names & todo_names
            this_todo_names |= hop2_names
            todo_names ^= hop2_names

        if len(comp_names) >= 2:
            components += [comp_names]
        else:
            name = comp_names.pop()
            new_name_mapping[name] = name_mapping[name]

    # Only components (i.e. name sets of >= 2 names) need further processing
    for comp_names in components:
        did_something = True
        chosen_name = id_preference(comp_names)
        for name in comp_names:
            # print("CHOSEN NAME COMPARE:", name, chosen_name, bool(name == chosen_name), bool(name != chosen_name))
            if name != chosen_name:
                relabelled_names[name] = chosen_name

        all_cmds = {cmd for name in comp_names for cmd in name_mapping[name]}
        new_name_mapping[chosen_name] = all_cmds

    # Now for each name, validate and recombine the commands into the smallest set.
    #
    # This means validating against every pair. And reducing every pair.
    #
    # Can reduce with every pair and expect either 1 or 0 outputs. 1 output
    # means it is a reduction of the pair and an output of None means that
    # the pair were distinct from one another. Any errors will just be
    # raised as exceptions.

    # print()
    # print("NEW NAME MAPPING")
    # for name,cmds in new_name_mapping.items():
    #     print(name)
    #     for cmd in cmds:
    #         print("  -", cmd)
    # print("---")

    # This might not be the best way to do it, but it will work and I can
    # think about how else to approach this.
    all_cmds = []

    for name in new_name_mapping:
        todo_cmds = set(new_name_mapping[name])
        done_cmds = []
        while len(todo_cmds) > 0:
            new_cmd = todo_cmds.pop()
            for i in range(len(done_cmds)):
                keep,new_cmds,later_cmds = recombination(done_cmds[i], new_cmd)
                if len(new_cmds) > 0:
                    did_something = True
                    todo_cmds |= set(new_cmds)
                if len(later_cmds) > 0:
                    did_something = True
                    # These will be handled in the next iteration
                    all_cmds += later_cmds
                if not keep:
                    did_something = True
                    done_cmds.pop(i)
                    break
            else:
                done_cmds += [new_cmd]
        all_cmds += done_cmds

    # print()
    # print("ALL CMDS")
    # for cmd in all_cmds:
    #     print(cmd)
    # print("---")


    # print()
    # print("NAMES")
    # print("=====")
    # print(relabelled_names)
    # print()

    # It's possible that we did something inadventently by the implicit unique in set.
    if len(all_cmds) != len(cmds_in):
        did_something = True

    return all_cmds, relabelled_names, did_something


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import report_import
report_import("zef.core.graph_additions.command_multi_rules")

from .common import *

AliasDict = Dict[AllIDs][AllIDs]

################################################################
# * Entrypoint dispatch functions
#--------------------------------------------------------------

def cull_unnecessary_cmds(cmds: List[PleaseCommandLevel1], gs: GraphSlice):
    # Remove cmds that should have no effect on the final graph. Becvause
    # culling could also remove ids, we should make sure that all aliasing of
    # names has been resolved before this point. Cmds that would create a link
    # between names (e.g. an instantiate that has an origin_uid and an internal
    # id) should raise an error at this point, as it indicates aliasing was not
    # properly resolved.
    #
    # I immediately ran into an issue with this for things like value nodes, or
    # delegates. So instead, the culling is allowed to produce new aliasing,
    # from IDs to ZefRefs.

    out_cmds,aliases = (
        cmds
        | map[lambda cmd: (cmd,gs) | match_rules[[
            *cull_rules,
            (Any, not_implemented_error["Unknown cmd type for culling"]),
        ]] | collect]
        # Checking that each cmd produces only 0-1 cmds
        | map[Assert[first | length | less_than_or_equal[1]]]
        | reducemany[concat, merge_nodups][[],{}]
        | collect)

    return out_cmds,aliases
        

def relabel_cmds(cmds: List[PleaseCommandLevel1], aliases: AliasDict) -> List[PleaseCommandLevel1]:
    out_cmds = (
        cmds
        | map[lambda cmd: (cmd,aliases) | match_rules[[
            *relabel_rules,
            (Any, not_implemented_error["Unknown cmd type for relabelling"]),
        ]] | collect]
        | collect)

    return out_cmds

def distinguish_cmd(cmd: PleaseCommandLevel1):
    # Returns the distinguishable representation of an object, without any other context known.
    #
    # Entity(ET.X, uid="abcd")[V.name] -> Entity(ET.X, uid="abcd")
    # ET.Machine[V.name] -> ET.Machine[InternalID(n+1)]
    # 
    # This is used to fill in the list of unique operations, which will be pointed at by other names later on.

    cmd_distinguish,names = cmd | match[[
        *distinguish_rules,
        (Any, not_implemented_error["Unknown cmd type"]),
    ]] | collect

    return cmd_distinguish,names

#########
# ** PleaseInstantiate

def cull_instantiate(cmd: PleaseInstantiate, gs):
    maybe_ezr = None
    if "origin_uid" in cmd:
        from ..graph_slice import get_instance_rae
        maybe_ezr = find_rae_in_target(cmd.origin_uid, gs)
    elif isinstance(cmd.atom, PleaseInstantiateValueNode):
        maybe_ezr = find_value_node_in_target(cmd.atom, gs)
    elif isinstance(cmd.atom, PleaseInstantiateDelegate):
        maybe_ezr = to_delegate(cmd.atom, gs, False)

    if maybe_ezr is not None:
        # We have an existing item, don't need to action this but do store the
        # found item for looking up wish ids.
        alias = {}
        if "internal_ids" in cmd:
            for wish_id in cmd.internal_ids:
                alias[wish_id] = maybe_ezr
            
        # We can discard this command.
        return [], alias

    return [cmd], {}

def relabel_instantiate(cmd: PleaseInstantiate, aliases: AliasDict) -> PleaseInstantiate:
    atom = cmd.atom
    if isinstance(atom, PleaseInstantiateRelation):
        atom = (atom
                | insert["source"][lookup_alias(atom["source"], aliases)]
                | insert["target"][lookup_alias(atom["target"], aliases)]
                | collect)

    d = dict(cmd._value)
    d = d | insert["atom"][atom] | collect
    if "internal_ids" in d:
        d = (d
             | insert["internal_ids"][cmd.internal_ids
                                      | map[lambda x: lookup_alias(x, aliases)]
                                      | collect]
             | collect)

    return PleaseInstantiate(d)


def distinguish_instantiate(cmd: PleaseInstantiate) -> Tuple[PleaseInstantiate, List[AllIDs]]:
    if isinstance(cmd.atom, PleaseInstantiateDelegate | PleaseInstantiateValueNode):
        # Delegate is special - its own type is its id
        names = [cmd.atom] + cmd._value.get("internal_ids", [])
        new_cmd = cmd._get_type()(cmd._value | without[["internal_ids"]] | collect)
        return new_cmd,names
    elif "origin_uid" in cmd:
        names = [cmd.origin_uid] + cmd._value.get("internal_ids", [])
        new_cmd = cmd._get_type()(cmd._value | without[["internal_ids"]] | collect)
        return new_cmd,names
    elif "internal_ids" in cmd and len(cmd.internal_ids) >= 1:
        names = cmd.internal_ids
        # Just use the first name as the unique name
        new_cmd = cmd._get_type()(cmd._value | insert["internal_ids"][[names[0]]] | collect)
        return new_cmd,names
    else:
        new_name = global_gen_internal_id()
        new_cmd = cmd._get_type()(cmd._value | insert["internal_ids"][[new_name]] | collect)
        return new_cmd,[new_name]

#########
# ** PleaseAssign

def cull_assign(cmd: PleaseAssign, gs):
    if isinstance(cmd.target, EternalUID):
        if cmd.target in gs:
            z = gs | get[cmd.target] | collect
            if Val(value(z)) == cmd.value:
                # Only here are we sure we can skip this!
                return [], {cmd.target: z}

    return [cmd], {}

def distinguish_assign(cmd: PleaseAssign) -> Tuple[PleaseAssign, List[AllIDs]]:
    if isinstance(cmd.target, AttributeEntityRef):
        name = origin_uid(cmd.target)
    else:
        name = cmd.target
    return cmd, [name]


#########
# ** PleaseTerminate

def cull_terminate(cmd: PleaseTerminate, gs):
    # If not on the graph, don't need to terminate
    if isinstance(cmd.target, EternalUID):
        if cmd.target not in gs:
            return [], {cmd.target: None}
    return [cmd], {}

def relabel_terminate(cmd: PleaseTerminate, aliases: AliasDict) -> PleaseTerminate:
    new_target = lookup_alias(cmd.target, aliases)
    return PleaseTerminate(target=new_target)
    

#########
# ** PleaseAlias

def cull_alias(cmd: PleaseAlias, gs):
    if len(cmd.ids) >= 2:
        raise Exception(f"Got to the point of culling an alias which is still shared across two names. This shouldn't happend! {cmd}")
    return [], {}

def relabel_alias(cmd: PleaseAlias, aliases: AliasDict) -> PleaseAlias:
    ids = cmd.ids | map[lambda x: lookup_alias(x, aliases)] | distinct | collect
    return PleaseAlias(ids=ids)

def distinguish_alias(cmd):
    return cmd, cmd.ids

#########
# ** PleaseMustLive

def distinguish_must_live(cmd):
    return cmd, [cmd.target]

#########
# ** General rules

def cull_unconditionally(cmd, gs):
    return [], {}
    

def relabel_just_target(cmd, aliases: AliasDict):
    # Works on any please command that has a "target" field that needs reassigning and nothing else.
    d = dict(cmd._value)
    d = d | insert["target"][lookup_alias(d["target"], aliases)] | collect

    return cmd._get_type()(d)

def distinguish_has_target(cmd):
    return cmd, [cmd.target]



######################################
# * Listing of rules
#------------------------------------

cull_rules = [
    (PleaseInstantiate, cull_instantiate),
    (PleaseAssign, cull_assign),
    (PleaseAlias, cull_alias),
    (PleaseTerminate, cull_terminate),
    # These don't have to be culled, could just be no-ops at the low level. 
    (PleaseBeSource | PleaseBeTarget, cull_unconditionally),
    (PleaseMustLive, cull_unconditionally),
]

relabel_rules = [
    (PleaseInstantiate, relabel_instantiate),
    (PleaseAssign, relabel_just_target),
    (PleaseBeSource | PleaseBeTarget, relabel_just_target),
    (PleaseMustLive, relabel_just_target),
    (PleaseAlias, relabel_alias),
    (PleaseTerminate, relabel_terminate),
]
        
    
distinguish_rules = [
    (PleaseInstantiate, distinguish_instantiate),
    (PleaseAssign, distinguish_assign),
    (PleaseBeSource | PleaseBeTarget, distinguish_has_target),
    (PleaseAlias, distinguish_alias),
    (PleaseMustLive, distinguish_must_live),
    (PleaseTerminate, distinguish_has_target),
]


##############################
# * Utils
#----------------------------


def lookup_alias(name: AllIDs, aliases: AliasDict) -> AllIDs:
    # We have to do this recursively, in case there's an alias chain
    while name in aliases:
        # Just for santiy check at the moment.
        assert id_preference_pair(aliases[name], name) == aliases[name]
        name = aliases[name]
    return name

def resolved_variables_from_aliases(aliases: AliasDict) -> ResolvedVariableDict:
    out = {}
    for k in aliases:
        if isinstance(k, Variable):
            out[k] = lookup_alias(k, aliases)

    assert isinstance(out, ResolvedVariableDict)
    return out




# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from cogapp import Cog

import sys
search_path = sys.argv[1]
output_path = sys.argv[2]
tokens_path = sys.argv[3]

def find_files_of_type(path, filename_endings, directories_to_exclude={}):
    # Returns paths relative to the input path
    for dirpath,dirnames,filenames in os.walk(path):
        for filename in filenames:
            thispath = os.path.join(dirpath, filename)
            if os.path.splitext(thispath)[1] in filename_endings:
                yield os.path.relpath(thispath, path)

        for dirname in directories_to_exclude:
            while dirname in dirnames:
                dirnames.remove(dirname)

cog = Cog()
cog.options.bReplace = False
cog.options.bDeleteCode = True
cog.options.sPrologue = f"""
import cog
import json
import os

et_filename = os.path.join("{tokens_path}", "zeftypes_ET.json")
with open(et_filename) as F:
    et = json.loads(F.read())

rt_filename = os.path.join("{tokens_path}", "zeftypes_RT.json")
with open(rt_filename) as F:
    rt = json.loads(F.read())

kw_filename = os.path.join("{tokens_path}", "zeftypes_KW.json")
with open(kw_filename) as F:
    kw = json.loads(F.read())

en_filename = os.path.join("{tokens_path}", "zeftypes_EN.json")
with open(en_filename) as F:
    en = json.loads(F.read())

def enum_type(x):
    return x.split('.')[0]
def enum_val(x):
    return x.split('.')[1]

def is_c_keyword(s):
    return s in [
        "alignas",
        "alignof",
        "and",
        "and_eq",
        "asm",
        "atomic_cancel",
        "atomic_commit",
        "atomic_noexcept",
        "auto",
        "bitand",
        "bitor",
        "bool",
        "break",
        "case",
        "catch",
        "char",
        "char8_t",
        "char16_t",
        "char32_t",
        "class",
        "compl",
        "concept",
        "const",
        "consteval",
        "constexpr",
        "constinit",
        "const_cast",
        "continue",
        "co_await",
        "co_return",
        "co_yield",
        "decltype",
        "default",
        "delete",
        "do",
        "double",
        "dynamic_cast",
        "else",
        "enum",
        "explicit",
        "export",
        "extern",
        "false",
        "float",
        "for",
        "friend",
        "goto",
        "if",
        "inline",
        "int",
        "long",
        "mutable",
        "namespace",
        "new",
        "noexcept",
        "not",
        "not_eq",
        "nullptr",
        "operator",
        "or",
        "or_eq",
        "private",
        "protected",
        "public",
        "reflexpr",
        "register",
        "reinterpret_cast",
        "requires",
        "return",
        "short",
        "signed",
        "sizeof",
        "static",
        "static_assert",
        "static_cast",
        "struct",
        "switch",
        "synchronized",
        "template",
        "this",
        "thread_local",
        "throw",
        "true",
        "try",
        "typedef",
        "typeid",
        "typename",
        "union",
        "unsigned",
        "using",
        "virtual",
        "void",
        "volatile",
        "wchar_t",
        "while",
        "xor",
        "xor_eq",
    ]
"""


#cog.options.verbosity = 0
for filename in find_files_of_type(path=search_path, filename_endings={'.cog'}):
    try:
        true_output = os.path.join(output_path, filename[:-len(".cog")] + ".gen")
        os.makedirs(os.path.dirname(true_output), exist_ok=True)
        cog.options.sOutputName = true_output + ".tmp"
        cog.processOneFile(os.path.join(search_path, filename))
        if not os.path.exists(true_output) or open(true_output + ".tmp").read() != open(true_output).read():
            print(filename, " changed")
            os.rename(true_output + ".tmp", true_output)
        else:
            os.unlink(true_output + ".tmp")
    except Exception as exc:
        print(f'An exception was raised when processing file "{filename}": {exc}')
        # Need this to fail for cmake to not continue on without a care.
        raise



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import report_import
report_import("zef.core.graph_additions.command_recombination")

from .common import *


def recombination(cmd1, cmd2):
    # When two objects have been identified as being the same, this is the procedure (or problem-aware point) for combining the two objects together.
    #
    # (ET.Machine[InternalID(1)], ET.Machine[InternalID(2)]) -> ET.Machine[InternalID(n+1)]
    # (Entity(ET.Machine, uid="abcd"), ET.Machine[InternalID(1)]) -> Entity(ET.Machine, uid="abcd")
    # (ET.X[InternalID(1)], ET.Y[InternalID(2)]) -> ERROR

    # print("Doing recomb for", cmd1, cmd2)
    return (cmd1, cmd2) | match[[
        *recombination_rules,
        (Any, not_implemented_error["Unknown cmd pair type"]),
    ]] | collect
    
def recombination_unchanged(_):
    return True,[],[]

def recombination_invalid(arg):
    x,y = arg
    raise Exception(f"Can't have two commands for the same name which contradict one another: {x} and {y}")

def recombination_checks(cmds):
    cmd1,cmd2 = cmds
    
    def get_target(x):
        if isinstance(x, PleaseInstantiate):
            return x._value.get("origin_uid", None)
        else:
            if isinstance(x.target, AtomRef):
                return x.target
            return None
    target1 = get_target(cmd1)
    target2 = get_target(cmd2)

    if target1 is not None and target2 is not None:
        if target1 != target2:
            raise Exception("Two commands refer to different origin uids but have the same name")
        target = target1
    elif target1 is None:
        target = target2
    else:
        target = target1

    if cmd1._get_type() == cmd2._get_type():
        if isinstance(cmd1, PleaseInstantiate):
            if target is not None:
                # At least one of the commands has an origin uid - we take this version
                if "origin_uid" in cmd1:
                    new_cmd = cmd1
                else:
                    new_cmd = cmd2
            else:
                # Neither command has an origin uid - so just take the first one (will be a given name)
                new_cmd = cmd1

            new_cmds = [new_cmd]
            later_cmds = []

            if isinstance(cmd1.atom, PleaseInstantiateRelation):
                assert isinstance(cmd2.atom, PleaseInstantiateRelation)
                assert cmd1.atom["rt"] == cmd2.atom["rt"]
                if cmd1.atom["source"] != cmd2.atom["source"]:
                    later_cmds += [PleaseAlias(ids=[force_as_id(cmd1.atom["source"]), force_as_id(cmd2.atom["source"])])]
                if cmd1.atom["target"] != cmd2.atom["target"]:
                    later_cmds += [PleaseAlias(ids=[force_as_id(cmd1.atom["target"]), force_as_id(cmd2.atom["target"])])]
            else:
                assert cmd1.atom == cmd2.atom, f"Atoms are different between two instantiations: {cmd1}, {cmd2}"

            return False, new_cmds, later_cmds
        elif isinstance(cmd1, PleaseAssign):
            if cmd1.value != cmd2.value:
                raise Exception(f"Two assigns have different values: {cmd1} {cmd2}")
            # Any of the ids will do - they will be relabelled anyway.
            _,names = distinguish_assign(cmd1)
            return False, [PleaseAssign(target=names[0], value=cmd1.value)], []
        else:
            raise NotImplementedError(f"TODO: {cmd1._get_type()}")
    else:
        # Two different commands, just leave as is
        return True, [], []

def recombination_be_source_target(cmds):
    cmd1,cmd2 = cmds

    if not cmd1.exact and not cmd2.exact:
        return True, [], []

    # Since either of these are exact, we must be able to combine these together
    # to be the same. Ideally we would be able to account for multiple BeSource
    # or BeTargets at once, but since we're doing this one at a time, we just
    # have to make sure:
    # a) Any non-exact is a subset of any exact
    # b) Two exacts are exactly the same
    #
    # However, note that a subset is tricky - which of the rels is matched up to
    # other rels? So will fail in that case.
    #
    # We'll try and match with sets but this is going to be impossible locally,
    # it requires global information. For now, we'll put them in as if they are
    # ordered lists which will work if the user passes in identical sets twice.
    if cmd1.rt == cmd2.rt:
        if len(set(cmd1.rel_ids) - set(cmd2.rel_ids)) == 0:
            # cmd2 is a subset
            return False, [cmd1], []
        if len(set(cmd2.rel_ids) - set(cmd1.rel_ids)) == 0:
            # cmd1 is a subset
            return False, [cmd2], []

        if len(cmd1.rel_ids) != len(cmd2.rel_ids):
            raise Exception("Can't match up ids blindly")

        # Going to match up ids blindy and hope for the best
        # Note: using lists to preserve order
        alias_cmds = []
        cmd1_diff = [id for id in cmd1.rel_ids if id not in cmd2.rel_ids]
        cmd2_diff = [id for id in cmd2.rel_ids if id not in cmd1.rel_ids]
        for id1,id2 in zip(cmd1_diff, cmd2_diff):
            other_cmds += [PleaseAlias(ids=[cmd1.rel_id, cmd2.rel_id])]
        return False, [cmd1], alias_cmds

    return True, [], []

def recombination_alias(cmds):
    cmd1,cmd2 = cmds

    return False, [PleaseAlias(ids=distinct(cmd1.ids + cmd2.ids))], []
            

recombination_rules = [
    (Tuple[PleaseBeSource, PleaseBeSource], recombination_be_source_target),
    (Tuple[PleaseBeTarget, PleaseBeTarget], recombination_be_source_target),
    # (Tuple[PleaseBeSource, ~PleaseBeSource], recombination_pass_through),
    # (Tuple[PleaseBeTarget, ~PleaseBeTarget], recombination_pass_through),
    (Tuple[PleaseAlias, PleaseAlias], recombination_alias),
    (Tuple[PleaseAlias, ~PleaseAlias], recombination_unchanged),
    (Tuple[~PleaseAlias, PleaseAlias], recombination_unchanged),
    (Tuple[PleaseMustLive, PleaseMustLive], first | wrap_list),
    (Tuple[~PleaseTerminate, ~PleaseTerminate], recombination_checks),
    (Tuple[~PleaseTerminate, PleaseTerminate], recombination_invalid),
    (Tuple[PleaseTerminate, ~PleaseTerminate], recombination_invalid),
]




# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import report_import
report_import("zef.core.graph_additions.wish_interpretation")

from .common import *
from .wish_tagging import Taggable, ensure_tag

##########################################
# * Level 2 generation
#----------------------------------------

Lvl2Context = Pattern[{"gen_id_state": GenIDState,
                       Optional["custom"]: Any}]

# Rules should follow the signature below, but we can't specify that just yet.
# TODO: maybe the element types should be Anys so that rules can be extended?
#InterpretationFunc = Func[GraphWishInput, Lvl2Context][List[PleaseCommandLevel2], List[GraphWishInput], Lvl2Context]
InterpretationFunc = Any


# Really want to use Subtype[GraphWishInput] instead of ValueType
def generate_level2_commands(inputs: List[GraphWishInput], rules: List[Tuple[ValueType,InterpretationFunc]]):
    # For each input
    #
    # - if it's a please command, just include
    # - if it's a simple translatable one (object notation, value, triple)
    #   - convert to sequence of commands
    # - if its a zefop, then store as PleaseRun

    output_cmds = []
    context = {
        "gen_id_state": generate_initial_state("lvl2"),
    }

    todo = list(inputs)
    while len(todo) > 0:
        input = todo.pop(0)
        cmds,new_inputs,context = (input,context) | match_rules[[
            *rules,
            (Any, not_implemented_error["Unknown type of object as input to graph wish"]),
        ]] | collect
        output_cmds += cmds
        todo = new_inputs + todo
        # print()
        # print("======")
        # print(input)
        # print(cmds)
        # print(new_inputs)
        # print()

    output = {
        "cmds": output_cmds,
    }
    if "custom" in context:
        output["custom"] = context["custom"]
    return output

@func
def lvl2cmds_for_ETorAET(input: ET | AET, context: Lvl2Context):
    names = names_of_raet(input)
    this_bare_raet = bare_raet(input)

    d = {"atom": this_bare_raet}

    ouids = set(names | filter[EternalUID] | collect)
    iids = set(names | filter[Not[EternalUID]] | collect)

    if len(ouids) >= 2:
        raise Exception(f"Can't have multiple origin uids for the one object: {ouids}")
    if len(ouids) == 1:
        d["origin_uid"] = single(ouids)

    if len(iids) > 0:
        d["internal_ids"] = tuple(iids)

    return [PleaseInstantiate(d)], [], context

@func
def lvl2cmds_for_delegate(input: Delegate, context: Lvl2Context):
    d = to_delegate(input)
    cmd = PleaseInstantiate({"atom": d})

    return [cmd], [], context

@func
def lvl2cmds_for_valuenode(input: WrappedValue, context: Lvl2Context):
    cmd = PleaseInstantiate({"atom": input})

    return [cmd], [], context

@func
def lvl2cmds_for_relation_triple(input: RelationTriple, context: Lvl2Context):
    names = names_of_raet(input[1])
    bare_rt = bare_raet(input[1])

    gen_id_state = context["gen_id_state"]

    src_obj, src, gen_id_state = ensure_tag(input[0], gen_id_state)
    trg_obj, trg, gen_id_state = ensure_tag(input[2], gen_id_state)

    d = {
        "atom": {
            "rt": bare_rt,
            "source": src,
            "target": trg,
        }
    }
    euids = [x for x in names if isinstance(x, EternalUID)]
    others = [x for x in names if not isinstance(x, EternalUID)]
    if len(euids) > 0:
        assert len(euids) == 1
        d["origin_uid"] = euids[0]
    if len(others) > 0:
        d["internal_ids"] = others

    cmds = [PleaseInstantiate(d)]

    # If the RT is itself an atom with further fields, we need to pass those fields through too
    if isinstance(input[1], RelationAtom):
        cmds += [input[1]]

    context = context | insert["gen_id_state"][gen_id_state] | collect
    return cmds, [src_obj, trg_obj], context

@func
def OS_lvl2cmds_for_relation_triple(input: OldStyleRelationTriple, context: Lvl2Context):
    gen_id_state = context["gen_id_state"]

    more_inputs = []

    if len(input) == 2:
        s, rels = input
        src_obj, src, gen_id_state = ensure_tag(s, gen_id_state)
        more_inputs += [src_obj]

        for rt,t in rels:
            trg_obj, trg, gen_id_state = ensure_tag(t, gen_id_state)
            more_inputs += [trg_obj]
            more_inputs += [(src, rt, trg)]

    elif len(input) == 3:
        if isinstance(input[0], List):
            sources = input[0]
        else:
            sources = [input[0]]

        if isinstance(input[2], List):
            targets = input[2]
        else:
            targets = [input[2]]
        rt = input[1]

        src_names = []
        for src in sources:
            src_obj, src, gen_id_state = ensure_tag(src, gen_id_state)
            more_inputs += [src_obj]
            src_names += [src]
        trg_names = []
        for trg in targets:
            trg_obj, trg, gen_id_state = ensure_tag(trg, gen_id_state)
            more_inputs += [trg_obj]
            trg_names += [trg]

        more_inputs += [(s,rt,t) for s in src_names for t in trg_names]

    context = context | insert["gen_id_state"][gen_id_state] | collect
    return [], more_inputs, context

@func
def lvl2cmds_for_aet_with_value(input: AETWithValue, context: Lvl2Context):
    gen_id_state = context["gen_id_state"]
    input, me, gen_id_state = ensure_tag(input, gen_id_state)

    lvl2_cmds = [PleaseInstantiate({"atom": input.aet,
                                    "internal_ids": input.internal_ids}),
                 PleaseAssign({"target": me,
                               "value": input.value})]
            
    return lvl2_cmds, [], context

# @func
# Note: can't be a func, can't allow this to participate in the normal
# evaluation engine
def lvl2cmds_for_lazyvalue(input: LazyValue, context: Lvl2Context):
    from .obj_evaluation import evaluate_chain

    lvl2cmds = []
    further_cmds = []
    if isinstance(input.initial_val, AtomClass | Variable):
        obj = input.initial_val
        gen_id_state = context["gen_id_state"]
        final_state = evaluate_chain(obj, input.el_ops, gen_id_state)
        gen_id_state = final_state["gen_id_state"]
        further_cmds += final_state["emitted_cmds"]
        if isinstance(final_state["obj"], Nil | Variable):
            pass
        else:
            if final_state["obj"] == input:
                raise Exception("lvl2cmds is getting stuck in infinite loop when evaluating LazyValue")
            further_cmds += [final_state["obj"]]
    elif isinstance(input.initial_val, BlobPtr):
        obj = Atom(input.initial_val)
        new_lv = obj | input.el_ops
        further_cmds += [new_lv]
    elif isinstance(input.initial_val, FlatRef):
        obj = Atom(input.initial_val)
        new_lv = obj | input.el_ops
        further_cmds += [new_lv]
    elif isinstance(input.initial_val, RAERef):
        obj = Atom(input.initial_val)
        new_lv = obj | input.el_ops
        further_cmds += [new_lv]
    elif isinstance(input.initial_val, ET):
        obj = input.initial_val
        new_lv = obj() | input.el_ops
        further_cmds += [new_lv]
    elif isinstance(input.initial_val, ExtraUserAllowedIDs):
        obj = convert_extra_allowed_id(input.initial_val)
        new_lv = obj | input.el_ops
        further_cmds += [new_lv]
    else:
        raise Exception(f"Don't know what to do with an initial value of {input.initial_val}")

    return lvl2cmds, further_cmds, context


def lvl2cmds_for_rae(input: RAE, context: Lvl2Context):
    if isinstance(input, Entity|AttributeEntity):
        cmd = PleaseInstantiate(atom=rae_type(input),
                                origin_uid=origin_uid(input))
        return [cmd], [], context
    elif isinstance(input, Relation):
        gen_id_state = context["gen_id_state"]
        s_obj,s_name,gen_id_state = ensure_tag(source(input), gen_id_state)
        t_obj,t_name,gen_id_state = ensure_tag(target(input), gen_id_state)
        cmd = PleaseInstantiate(atom=dict(rt=rae_type(input),
                                          source=s_name,
                                          target=t_name),
                                origin_uid=origin_uid(input))
        context = context | insert["gen_id_state"][gen_id_state] | collect
        return [cmd], [s_obj,t_obj], context
    else:
        raise Exception("Shouldn't get here")

def lvl2cmds_for_flatref(input: FlatRef, context: Lvl2Context):
    return [], [Atom(input)], context

def lvl2cmds_for_flatrefs(input: FlatRefs, context: Lvl2Context):
    return [], list(input), context

def lvl2cmds_for_flatgraph(input: FlatGraph, context: Lvl2Context):
    new_cmds = list(input | all | collect)
    return [], new_cmds, context

@func
def lvl2cmds_for_se(input, context):
    # Note: this should not hit Variables as they will be handled in AllIDs before this
    assert not isinstance(input, Variable)

    from .obj_evaluation import cast_symbolic_expression_as_lazyvalue
    new_input = cast_symbolic_expression_as_lazyvalue(input)
    return [], [new_input], context
        

def discard_unnecessary_frame(x):
    # This function exists solely to handle Atoms with FlatRefs in them. The
    # problem with these is the presence of the FlatGraph is necessary to
    # uniquely identify them. Other Atoms can have that aspect removed.
    authority = get_most_authorative_id(x)
    if isinstance(authority, FlatRefUID):
        return x
    return discard_frame(x)
        
@func
def lvl2cmds_for_atom(atom, context):
    if isinstance(rae_type(atom), Nil):
        return [atom], [], context

    if isinstance(atom, Entity | TX | Root):
        return [discard_unnecessary_frame(atom)], [], context

    if isinstance(atom, Relation):
        # We need to ensure the relation source/target is also included.

        # If the relation atom is included without an id, either global or
        # local, then this is an error. If the atom were in a relation triple,
        # then we should have a local name by this point.
        authorative_id = get_most_authorative_id(atom)
        if authorative_id is None:
            raise Exception("Can't have relation Atom without source/target")

        if isinstance(authorative_id, WishID):
            # Good enough, source/target should be defined elsewhere
            return [atom], [], context

        # This means the atom should have a source/target. Let's find them
        # from the atom internals.

        z = find_concrete_pointer(atom)
        if z is None:
            raise Exception("Can't determine relation source/target because we can't load a concrete ZefRef for this Atom")

        src = Atom(source(z))
        trg = Atom(target(z))
        plain_rt = rae_type(atom)[authorative_id]

        return [discard_unnecessary_frame(atom)], [(src, plain_rt, trg)], context

    if isinstance(atom, AttributeEntity):
        # We need to ensure the value is also included, if the reference frame is accessible.
        #
        # If the frame is not given, we can assume "no value" is what is intended.

        authorative_id = get_most_authorative_id(atom)
        if authorative_id is None:
            return [atom], [], context

        if isinstance(authorative_id, WishID):
            # Good enough, value might be defined elsewhere
            return [atom], [], context

        from ..atom import _get_atom_id
        atom_id = _get_atom_id(atom)
        if "frame_uid" in atom_id:
            z = find_concrete_pointer(atom)
            if z is None:
                raise Exception("Can't determine value of AttributeEntity as frame is not loaded")
            val = value(z)
            if val is not None:
                cmds = [PleaseAssign(target=authorative_id, value=Val(val))]
            else:
                cmds = []
            return [discard_unnecessary_frame(atom)], cmds, context
        else:
            return [atom], [], context

    raise Exception(f"Shouldn't get here: {atom}")



@func
def OS_lvl2cmds_for_dict(input: OldStyleDict, context: Lvl2Context):
    # This must be of the form:
    # {ET.Something: {RT.One: 5, RT.Two: z}}
    assert isinstance(input, OldStyleDict)
    assert len(input) == 1

    gen_id_state = context["gen_id_state"]

    res = []

    source_input, rel_dict = single(input.items())

    source_input, me, gen_id_state = ensure_tag(source_input, gen_id_state)
    res += [source_input]

    for k,v in rel_dict.items():
        v,v_id,gen_id_state = ensure_tag(v, gen_id_state)
        res += [v]
        res += [(me, k, v_id)]

    context = context | insert["gen_id_state"][gen_id_state] | collect
    return [], res, context

@func
def pass_through(input, context):
    return [input], [], context

@func
def lvl2cmds_ignore(input, context):
    return [], [], context

default_interpretation_rules = [
    # Note that delegates and WrappedValues are included in IDs so we have to put this first.
    (Delegate, lvl2cmds_for_delegate),
    (WrappedValue, lvl2cmds_for_valuenode),
    # Ignore IDs as many commands are likely to create these
    (AllIDs, lvl2cmds_ignore),
    #
    (LazyValue, lvl2cmds_for_lazyvalue),
    (SymbolicExpression, lvl2cmds_for_se),
    # (ObjectNotation, pass_through),
    (AtomClass, lvl2cmds_for_atom),
    (PleaseCommandLevel1, pass_through),
    # (PleaseCommandLevel2, not_implemented_error["TODO"]),
    (PleaseCommandLevel2, pass_through),
    (PureET|PureAET, lvl2cmds_for_ETorAET),
    (RelationTriple, lvl2cmds_for_relation_triple),
    (AETWithValue, lvl2cmds_for_aet_with_value),
    (RAERef | RAEConcrete, lvl2cmds_for_rae),
    (FlatRef, lvl2cmds_for_flatref),
    (FlatRefs, lvl2cmds_for_flatrefs),
    (FlatGraph, lvl2cmds_for_flatgraph),
    # Things not in base version
    (OldStyleDict, OS_lvl2cmds_for_dict),
    (OldStyleRelationTriple, OS_lvl2cmds_for_relation_triple),
]



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import report_import
report_import("zef.core.graph_additions.wish_translation_oldstyle")

from ..VT import *
from .common import *
from ..symbolic_expression import V

from ..zef_functions import func

from . import wish_interpretation
from .wish_interpretation import Lvl2Context

from ..VT.rae_types import RAET_get_names

# NOTE: everything in this file uses "OS" to mean "old style"

# This is what we will use to interface with the newer style and map variables
# back and forth.
oldstyle_prefix = "__oldstyle_"

# These rules try to mimic the behaviour of graph_delta.py

def gen_mapping_id(prev_state: GenIDState) -> Tuple[Variable]:
    prefix,prev_id = prev_state
    next_id = prev_id + 1
    id_str = getattr(V, f"{oldstyle_prefix}{prefix}_{next_id}")
    next_state = (prefix, next_id)
    return id_str, next_state

def OS_names_of_raet(context, raet):
    # Anything that's not a ValueType is a name - integers, strings, Vs, etc...
    names = RAET_get_names(raet)
    # We map these to Vs, so we can pass them to the rest of the framework
    V_names = []
    for name in names:
        context,V_name = OS_add_name_mapping(context, name)
        V_names += [V_name]

    return context, V_names

def OS_bare_raet(raet):
    return raet._replace(absorbed=(RAET_get_token(raet),))

def OS_add_name_mapping(context, name):
    if isinstance(name, WishID):
        return context, name

    # Using mutations for speed, could change this later. If made immutable,
    # need to make sure the context dict is recreated with these new items.
    custom = context.setdefault("custom", {})
    mapping = custom.setdefault("name_mapping", {})

    if name in mapping:
        return context, mapping[name]

    gen_id_state = context["gen_id_state"]
    id,gen_id_state = gen_mapping_id(gen_id_state)

    mapping[name] = id

    context = context | insert["gen_id_state"][gen_id_state] | collect

    return context, id


@func
def OS_lvl2cmds_for_ETorAET(input: ET|AET, context: Lvl2Context):
    context, names = OS_names_of_raet(context, input)
    bare_raet = OS_bare_raet(input)

    d = {"atom": bare_raet}
    if len(names) > 0:
        d["internal_ids"] = names
    return [PleaseInstantiate(d)], [], context

@func
def OS_lvl2cmds_for_relation_triple(input: RelationTriple, context: Lvl2Context):
    print("Deprecation warning: graph wishes using dictionary inputs will be disallowed in the future.")

    context, names = OS_names_of_raet(context, input[1])
    bare_rt = OS_bare_raet(input[1])

    d = {
        "atom": {
            "rt": bare_rt,
            "source": input[0],
            "target": input[2],
        }
    }
    if len(names) > 0:
        d["internal_ids"] = names

    return [PleaseInstantiate(d)], [], context


# This is to be replaced with a way to be self-referential
OSTaggable_Cheat = Any
DictSyntax = Dict[PureET][Dict[PureRT][OSTaggable_Cheat]] & Is[length | equals[1]]
OS_Taggable = wish_interpretation.Taggable | DictSyntax

@func
def OS_lvl2cmds_for_dict(input: DictSyntax, context: Lvl2Context):
    # This must be of the form:
    # {ET.Something: {RT.One: 5, RT.Two: z}}
    assert isinstance(input, DictSyntax)
    assert len(input) == 1

    gen_id_state = context["gen_id_state"]

    res = []

    source_input, rel_dict = single(input.items())

    source_input, me, gen_id_state = OS_ensure_tag(source_input, gen_id_state)
    res += [source_input]

    for k,v in rel_dict.items():
        v,v_id,gen_id_state = OS_ensure_tag(v, gen_id_state)
        res += [v]
        res += [(me, k, v_id)]

    context = context | insert["gen_id_state"][gen_id_state] | collect
    return [], res, context
        

def OS_ensure_tag(obj: OS_Taggable, gen_id_state: GenIDState) -> Tuple[OS_Taggable, WishID, GenIDState]:
    if isinstance(obj, DictSyntax):
        main_obj = single(obj.keys())
        new_main_obj,me,gen_id_state = OS_ensure_tag(main_obj, gen_id_state)
        if new_main_obj != main_obj:
            obj = {new_main_obj: single(obj.values())}
    else:
        # Fallback to regular version
        obj,me,gen_id_state = wish_interpretation.ensure_tag(obj, gen_id_state)

    return obj,me,gen_id_state



OS_interpretation_rules = [
    # Unaffected
    (AETWithValue, wish_interpretation.lvl2cmds_for_aet_with_value),
    (ObjectNotation, wish_interpretation.lvl2cmds_for_object_notation),
    (PleaseCommandLevel1, wish_interpretation.pass_through),
    (PleaseCommandLevel2, not_implemented_error["TODO"]),
    # Slight changes
    (PureET | PureAET, OS_lvl2cmds_for_ETorAET),
    (RelationTriple, OS_lvl2cmds_for_relation_triple),
    # Things not in base version
    (Dict, OS_lvl2cmds_for_dict),
]



def OS_post_transact_rule(receipt: GraphWishReceipt, lvl2_custom: Dict[Any][WishID]):
    if lvl2_custom is None:
        return receipt
    print("Deprecation warning: graph wishes using names other than `V.x` variables will be disallowed in future versions.")

    # Translate any oldstyle variable names into what they were originally
    new_items = {}
    to_remove = []
    flipped_mapping = {val: key for key,val in lvl2_custom["name_mapping"].items()}
    for key in receipt:
        if key.name.startswith(oldstyle_prefix):
            to_remove += [key]
            new_items[flipped_mapping[key]] = receipt[key]

    return receipt | without[to_remove] | merge[new_items] | collect
            
    


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
from zef import *
from zef.ops import *

class MyTestCase(unittest.TestCase):
    def test_pattern(self):
        d = {"a": 1,
             "b": "string"}

        self.assertIsInstance(d, Pattern[{"a": Int,
                                          "b": String}])

        self.assertIsInstance(d, Pattern[{"a": Int | String,
                                          "b": Int | String}])

        self.assertIsInstance(d, Pattern[{"a": Int | String}])
        self.assertIsInstance(d, Pattern[{"b": Any}])

        self.assertNotIsInstance(d, Pattern[{"c": Any}])
        self.assertNotIsInstance(d, Pattern[{"a": Any, "b": Any, "c": Any}])

        self.assertIsInstance(d, Pattern[{"a": Any, "b": Any, Optional["c"]: Any}])

        self.assertIsInstance(d, Pattern[{Optional["a"]: Any, Optional["b"]: Any, Optional["c"]: Any}])


if __name__ == '__main__':
    unittest.main()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import report_import
report_import("zef.core.pure_utils")


def get_in_pure(d, path, default_val):
    if len(path) == 0: return d
    if type(d) != dict: return default_val
    return get_in_pure(d.get(path[0], default_val), path[1:], default_val)


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import report_import
report_import("zef.core.graph_additions.wish_translation")

from .common import *

def ensure_tag_atom(obj, gen_id_state):
    me = get_most_authorative_id(obj)
    if me is not None:
        me2 = force_as_id(me)
        if me != me2:
            # A slightly weird syntax, as we need to first *remove* names.
            obj = obj.__replace__(atom_id={})
            obj = obj(me2)
            me = me2
        return obj,me,gen_id_state
        
    me,gen_id_state = gen_internal_id(gen_id_state)

    obj = obj(me)

    return obj,me,gen_id_state

def ensure_tag_primitive(obj, gen_id_state):
    obj = convert_scalar(obj)
    me,gen_id_state = gen_internal_id(gen_id_state)
    # Create new AETWithValue with the id included.
    obj = obj._get_type()(obj._value | insert["internal_ids"][[me]] | collect)
    return obj,me,gen_id_state

def ensure_tag_aet(obj: AETWithValue, gen_id_state):
    if "internal_ids" in obj and len(obj.internal_ids) > 0:
        me = obj.internal_ids[0]
    else:
        me,gen_id_state = gen_internal_id(gen_id_state)
        # Create new AETWithValue with the id included.
        obj = obj._get_type()(obj._value | insert["internal_ids"][[me]] | collect)
    return obj,me,gen_id_state

def ensure_tag_pure_et_aet(obj, gen_id_state):
    names = names_of_raet(obj)
    if len(names) == 0:
        me,gen_id_state = gen_internal_id(gen_id_state)
        obj = obj[me]
    else:
        me = names[0]
    return obj,me,gen_id_state

def ensure_tag_delegate(obj, gen_id_state):
    obj = to_delegate(obj)
    return obj,obj,gen_id_state

def ensure_tag_assign(obj: PleaseAssign, gen_id_state):
    return obj, force_as_id(obj.target), gen_id_state

def ensure_tag_OS_dict(obj: OldStyleDict, gen_id_state):
    main_obj = single(obj.keys())
    main_obj,obj_id,gen_id_state = ensure_tag(main_obj, gen_id_state)
    obj = {main_obj: single(obj.values())}
    return obj,obj_id,gen_id_state

def ensure_tag_rae_ref(obj: RAERef, gen_id_state):
    return obj,origin_uid(obj),gen_id_state

def ensure_tag_blob_ptr(obj: BlobPtr, gen_id_state):
    obj = discard_frame(obj)
    # BlobPtrs could be RAEs or other things like value nodes/delegates/txs, so
    # pass this back through to ensure_tag to dispatch on the right thing.
    return ensure_tag(obj, gen_id_state)
    # return obj,origin_uid(obj),gen_id_state

def ensure_tag_flatref(obj: FlatRef, gen_id_state):
    from ..flat_graph import FlatRef_maybe_uid
    fr_uid = FlatRef_maybe_uid(obj)
    if isinstance(fr_uid, EternalUID):
        me = fr_uid
    else:
        from ..atom import make_flatref_uid
        me = make_flatref_uid(obj)

    return obj,me,gen_id_state

def ensure_tag_pass_through(obj, gen_id_state):
    return obj,obj,gen_id_state

def ensure_tag_extra_user_id(obj: ExtraUserAllowedIDs, gen_id_state):
    obj = convert_extra_allowed_id(obj)
    return obj,obj,gen_id_state


tagging_rules = [
    (AtomClass, ensure_tag_atom),
    (PrimitiveValue, ensure_tag_primitive),
    (AETWithValue, ensure_tag_aet),
    (PureET | PureAET, ensure_tag_pure_et_aet),
    (Delegate, ensure_tag_delegate),
    (PleaseAssign, ensure_tag_assign),
    (RAERef, ensure_tag_rae_ref),
    (FlatRef, ensure_tag_flatref),
    (BlobPtr, ensure_tag_blob_ptr),
    (OldStyleDict, ensure_tag_OS_dict),
    (WrappedValue, ensure_tag_pass_through),
    (AllIDs, ensure_tag_pass_through),
    (ExtraUserAllowedIDs, ensure_tag_extra_user_id),
]
    
Taggable = Union[tagging_rules | map[first] | func[tuple] | collect]

def ensure_tag(obj: Taggable, gen_id_state: GenIDState) -> Tuple[Taggable, WishID, GenIDState]:

    return (obj, gen_id_state) | match_rules[[
        *tagging_rules,
        (Any, not_implemented_error["Don't know how to ensure a tag for object"]),
    ]] | collect